from typing import Optional, Union
from pydantic import BaseModel


class Document(BaseModel):
    text: str
    doc_id: Optional[Union[str, int]] = None
    metadata: Optional[dict] = None

    def __init__(
        self,
        text: str,
        doc_id: Optional[Union[str, int]] = None,
        metadata: Optional[dict] = None,
    ):
        super().__init__(text=text, doc_id=doc_id, metadata=metadata)



import pytest
from rerankers.results import Result, RankedResults
from rerankers.documents import Document


def test_ranked_results_functions():
    results = RankedResults(
        results=[
            Result(document=Document(id=0, text="Doc 0"), score=0.9, rank=2),
            Result(document=Document(id=1, text="Doc 1"), score=0.95, rank=1),
        ],
        query="Test Query",
        has_scores=True,
    )
    assert results.results_count() == 2
    top_k = results.top_k(1)
    assert len(top_k) == 1
    assert top_k[0].doc_id == 1
    assert results.get_score_by_docid(0) == 0.9


def test_result_attributes():
    result = Result(document=Document(id=1, text="Doc 1"), score=0.95, rank=1)
    assert result.doc_id == 1
    assert result.text == "Doc 1"
    assert result.score == 0.95
    assert result.rank == 1


def test_result_validation_error():
    with pytest.raises(ValueError) as excinfo:
        Result(doc_id=2, text="Doc 2")
    assert "Either score or rank must be provided." in str(excinfo.value)



from typing import Union, Optional, List, Iterable
from rerankers.documents import Document


def vprint(txt: str, verbose: int) -> None:
    if verbose > 0:
        print(txt)


try:
    import torch

    def get_dtype(
        dtype: Optional[Union[str, torch.dtype]],
        device: Optional[Union[str, torch.device]],
        verbose: int = 1,
    ) -> torch.dtype:
        if dtype is None:
            vprint("No dtype set", verbose)
        if device == "cpu":
            vprint("Device set to `cpu`, setting dtype to `float32`", verbose)
            dtype = torch.float32
        if not isinstance(dtype, torch.dtype):
            if dtype == "fp16" or "float16":
                dtype = torch.float16
            elif dtype == "bf16" or "bfloat16":
                dtype = torch.bfloat16
            else:
                dtype = torch.float32
        vprint(f"Using dtype {dtype}", verbose)
        return dtype

    def get_device(
        device: Optional[Union[str, torch.device]],
        verbose: int = 1,
        no_mps: bool = False,
    ) -> Union[str, torch.device]:
        if not device:
            vprint("No device set", verbose)
            if torch.cuda.is_available():
                device = "cuda"
            elif torch.backends.mps.is_available() and not no_mps:
                device = "mps"
            else:
                device = "cpu"
            vprint(f"Using device {device}", verbose)
        return device

except ImportError:
    print("Torch not installed...")


def make_documents(
    docs: List[str],
    doc_ids: Optional[Union[List[str], List[int]]] = None,
):
    if doc_ids is None:
        doc_ids = list(range(len(docs)))
    return [Document(doc, doc_id=doc_ids[i]) for i, doc in enumerate(docs)]


def prep_docs(
    docs: Union[str, List[str], Document, List[Document]],
    doc_ids: Optional[Union[List[str], List[int]]] = None,
    metadata: Optional[List[dict]] = None,
):
    if isinstance(docs, Document) or (
        isinstance(docs, List) and isinstance(docs[0], Document)
    ):
        if isinstance(docs, Document):
            docs = [docs]
        if doc_ids is not None:
            if docs[0].doc_id is not None:
                print(
                    "Overriding doc_ids passed within the Document objects with explicitely passed doc_ids!"
                )
                print(
                    "This is not the preferred way of doing so, please double-check your code."
                )
            for i, doc in enumerate(docs):
                doc.doc_id = doc_ids[i]

        elif doc_ids is None:
            doc_ids = [doc.doc_id for doc in docs]
            if doc_ids[0] is None:
                print(
                    "'None' doc_ids detected, reverting to auto-generated integer ids..."
                )
                doc_ids = list(range(len(docs)))

        if metadata is not None:
            if docs[0].meatadata is not None:
                print(
                    "Overriding doc_ids passed within the Document objects with explicitely passed doc_ids!"
                )
                print(
                    "This is not the preferred way of doing so, please double-check your code."
                )
            for i, doc in enumerate(docs):
                doc.metadata = metadata[i]

        return docs

    if isinstance(docs, str):
        docs = [docs]
    if doc_ids is None:
        doc_ids = list(range(len(docs)))
    if metadata is None:
        metadata = [{} for _ in docs]
    return [
        Document(doc, doc_id=doc_ids[i], metadata=metadata[i])
        for i, doc in enumerate(docs)
    ]


def get_chunks(iterable: Iterable, chunk_size: int):  # noqa: E741
    """
    Implementation from https://github.com/unicamp-dl/InRanker/blob/main/inranker/base.py with extra typing and more descriptive names.
    This method is used to split a list l into chunks of batch size n.
    """
    for i in range(0, len(iterable), chunk_size):
        yield iterable[i : i + chunk_size]



"""
Code for InRanker is taken from the excellent InRanker repo https://github.com/unicamp-dl/InRanker under its Apache 2.0 license.
The only change to the original implementation is the removal of InRanker's BaseRanker, replacing it with our own to support the unified API better.
The main purpose for adapting this code here rather than installing the InRanker library is to ensure greater version compatibility (InRanker requires Python >=3.10)
"""

from typing import List, Optional, Union
from math import ceil

import torch
from tqdm.auto import tqdm
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer
from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document


import torch

from rerankers.results import RankedResults, Result
from rerankers.utils import (
    vprint,
    get_device,
    get_dtype,
    prep_docs,
    get_chunks,
)

PREDICTION_TOKENS = {
    "default": ["▁false", "▁true"],
    "castorini/monot5-base-msmarco": ["▁false", "▁true"],
    "castorini/monot5-base-msmarco-10k": ["▁false", "▁true"],
    "castorini/monot5-large-msmarco": ["▁false", "▁true"],
    "castorini/monot5-large-msmarco-10k": ["▁false", "▁true"],
    "castorini/monot5-base-med-msmarco": ["▁false", "▁true"],
    "castorini/monot5-3b-med-msmarco": ["▁false", "▁true"],
    "castorini/monot5-3b-msmarco-10k": ["▁false", "▁true"],
    "unicamp-dl/InRanker-small": ["▁false", "▁true"],
    "unicamp-dl/InRanker-base": ["▁false", "▁true"],
    "unicamp-dl/InRanker-3B": ["▁false", "▁true"],
    "unicamp-dl/mt5-base-en-msmarco": ["▁no", "▁yes"],
    "unicamp-dl/ptt5-base-pt-msmarco-10k-v2": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-pt-msmarco-100k-v2": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-en-pt-msmarco-100k-v2": ["▁não", "▁sim"],
    "unicamp-dl/mt5-base-en-pt-msmarco-v2": ["▁no", "▁yes"],
    "unicamp-dl/mt5-base-mmarco-v2": ["▁no", "▁yes"],
    "unicamp-dl/mt5-base-en-pt-msmarco-v1": ["▁no", "▁yes"],
    "unicamp-dl/mt5-base-mmarco-v1": ["▁no", "▁yes"],
    "unicamp-dl/ptt5-base-pt-msmarco-10k-v1": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-pt-msmarco-100k-v1": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-en-pt-msmarco-10k-v1": ["▁não", "▁sim"],
    "unicamp-dl/mt5-3B-mmarco-en-pt": ["▁", "▁true"],
    "unicamp-dl/mt5-13b-mmarco-100k": ["▁", "▁true"],
    "unicamp-dl/monoptt5-small": ["▁Não", "▁Sim"],
    "unicamp-dl/monoptt5-base": ["▁Não", "▁Sim"],
    "unicamp-dl/monoptt5-large": ["▁Não", "▁Sim"],
    "unicamp-dl/monoptt5-3b": ["▁Não", "▁Sim"],
}


def _get_output_tokens(model_name_or_path, token_false: str, token_true: str):
    if token_false == "auto":
        if model_name_or_path in PREDICTION_TOKENS:
            token_false = PREDICTION_TOKENS[model_name_or_path][0]
        else:
            token_false = PREDICTION_TOKENS["default"][0]
            print(
                f"WARNING: Model {model_name_or_path} does not have known True/False tokens. Defaulting token_false to `{token_false}`."
            )
    if token_true == "auto":
        if model_name_or_path in PREDICTION_TOKENS:
            token_true = PREDICTION_TOKENS[model_name_or_path][1]
        else:
            token_true = PREDICTION_TOKENS["default"][1]
            print(
                f"WARNING: Model {model_name_or_path} does not have known True/False tokens. Defaulting token_true to `{token_true}`."
            )

    return token_false, token_true


class T5Ranker(BaseRanker):
    def __init__(
        self,
        model_name_or_path: str,
        batch_size: int = 32,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Optional[Union[str, torch.device]] = None,
        verbose: int = 1,
        token_false: str = "auto",
        token_true: str = "auto",
        return_logits: bool = False,
        inputs_template: str = "Query: {query} Document: {text} Relevant:"
    ):
        """
        Implementation of the key functions from https://github.com/unicamp-dl/InRanker/blob/main/inranker/rankers.py
        Changes are detailed in the docstring for each relevant function.

        T5Ranker is a wrapper for using Seq2Seq models for ranking.
        Args:
            batch_size: The batch size to use when encoding.
            dtype: Data type for model weights.
            device: The device to use for inference ("cpu", "cuda", or "mps").
            verbose: Verbosity level.
            silent: Whether to show progress bars.
        """
        self.verbose = verbose
        self.device = get_device(device, self.verbose, no_mps=True)
        self.dtype = get_dtype(dtype, self.device, self.verbose)
        self.batch_size = batch_size
        vprint(
            f"Loading model {model_name_or_path}, this might take a while...",
            self.verbose,
        )
        vprint(f"Using device {self.device}.", self.verbose)
        vprint(f"Using dtype {self.dtype}.", self.verbose)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(
            model_name_or_path, torch_dtype=self.dtype
        ).to(self.device)
        self.model.eval()
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)

        token_false, token_true = _get_output_tokens(
            model_name_or_path=model_name_or_path,
            token_false=token_false,
            token_true=token_true,
        )
        self.token_false_id = self.tokenizer.convert_tokens_to_ids(token_false)
        self.token_true_id = self.tokenizer.convert_tokens_to_ids(token_true)
        vprint(f"T5 true token set to {token_true}", self.verbose)
        vprint(f"T5 false token set to {token_false}", self.verbose)

        self.return_logits = return_logits
        if self.return_logits:
            vprint(
                f"Returning raw logits for `{token_true}` as scores...", self.verbose
            )
        else:
            vprint("Returning normalised scores...", self.verbose)
        self.inputs_template = inputs_template
        vprint(f"Inputs template set to {inputs_template}", self.verbose)

    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
    ) -> RankedResults:
        """
        Ranks a list of documents based on their relevance to the query.
        """
        docs = prep_docs(docs, doc_ids, metadata)
        scores = self._get_scores(query, [d.text for d in docs])
        ranked_results = [
            Result(document=doc, score=score, rank=idx + 1)
            for idx, (doc, score) in enumerate(
                sorted(zip(docs, scores), key=lambda x: x[1], reverse=True)
            )
        ]
        return RankedResults(results=ranked_results, query=query, has_scores=True)

    def score(self, query: str, doc: str) -> float:
        """
        Scores a single document's relevance to a query.
        """
        scores = self._get_scores(query, [doc])
        return scores[0] if scores else 0.0

    @torch.no_grad()
    def _get_scores(
        self,
        query: str,
        docs: List[str],
        max_length: int = 512,
        batch_size: Optional[int] = None,
    ) -> List[float]:
        """
        Implementation from https://github.com/unicamp-dl/InRanker/blob/main/inranker/rankers.py.
        Lightly modified so only the positive logits are returned and renamed the chunking function.

        Given a query and a list of documents, return a list of scores.
        Args:
            query: The query string.
            docs: A list of document strings.
            max_length: The maximum length of the input sequence.
        """
        if self.return_logits:
            logits = []
        else:
            scores = []
        if batch_size is None:
            batch_size = self.batch_size
        for batch in tqdm(
            get_chunks(docs, batch_size),
            disable=not self.verbose,
            desc="Scoring...",
            total=ceil(len(docs) / batch_size),
        ):
            queries_documents = [
                self.inputs_template.format(query=query, text=text)
                for text in batch
            ]
            tokenized = self.tokenizer(
                queries_documents,
                padding=True,
                truncation="longest_first",
                return_tensors="pt",
                max_length=max_length,
            ).to(self.device)
            input_ids = tokenized["input_ids"]
            attention_mask = tokenized["attention_mask"]

            _, batch_scores = self._greedy_decode(
                model=self.model,
                input_ids=input_ids,
                length=1,
                attention_mask=attention_mask,
                return_last_logits=True,
            )
            batch_scores = batch_scores[
                :, [self.token_false_id, self.token_true_id]
            ].cpu()
            if self.return_logits:
                logits.extend(batch_scores[:, 1].tolist())
            else:
                batch_scores = torch.log_softmax(batch_scores, dim=-1)
                batch_scores = torch.exp(batch_scores[:, 1])
                scores.extend(batch_scores.tolist())

        if self.return_logits:
            return logits
        return scores

    @torch.no_grad()
    def _greedy_decode(
        self,
        model,
        input_ids: torch.Tensor,
        length: int,
        attention_mask: torch.Tensor = None,
        return_last_logits: bool = True,
    ):
        """Implementation from https://github.com/unicamp-dl/InRanker/blob/main/inranker/rankers.py"""
        decode_ids = torch.full(
            (input_ids.size(0), 1),
            model.config.decoder_start_token_id,
            dtype=torch.long,
        ).to(input_ids.device)
        encoder_outputs = model.get_encoder()(input_ids, attention_mask=attention_mask)
        next_token_logits = None
        for _ in range(length):
            model_inputs = model.prepare_inputs_for_generation(
                decode_ids,
                encoder_outputs=encoder_outputs,
                past=None,
                attention_mask=attention_mask,
                use_cache=True,
            )
            outputs = model(**model_inputs)  # (batch_size, cur_len, vocab_size)
            next_token_logits = outputs[0][:, -1, :]  # (batch_size, vocab_size)
            decode_ids = torch.cat(
                [decode_ids, next_token_logits.max(1)[1].unsqueeze(-1)], dim=-1
            )
        if return_last_logits:
            return decode_ids, next_token_logits
        return decode_ids



from abc import ABC, abstractmethod
from asyncio import get_event_loop
from functools import partial
from typing import List, Optional, Union
from rerankers.results import RankedResults
from rerankers.documents import Document


class BaseRanker(ABC):
    @abstractmethod
    def __init__(self, model_name_or_path: str, verbose: int):
        pass

    @abstractmethod
    def score(self, query: str, doc: str) -> float:
        pass

    @abstractmethod
    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
    ) -> RankedResults:
        """
        End-to-end reranking of documents.
        """
        pass

    async def rank_async(
        self,
        query: str,
        docs: List[str],
        doc_ids: Optional[Union[List[str], str]] = None,
    ) -> RankedResults:


        loop = get_event_loop()
        return await loop.run_in_executor(None, partial(self.rank, query, docs, doc_ids))

    def as_langchain_compressor(self, k: int = 10):
        try:
            from rerankers.integrations.langchain import RerankerLangChainCompressor

            return RerankerLangChainCompressor(model=self, k=k)
        except ImportError:
            print(
                "You need to install langchain and langchain_core to export a reranker as a LangChainCompressor!"
            )
            print(
                'Please run `pip install "rerankers[langchain]"` to get all the required dependencies.'
            )



AVAILABLE_RANKERS = {}

try:
    from rerankers.models.transformer_ranker import TransformerRanker

    AVAILABLE_RANKERS["TransformerRanker"] = TransformerRanker
except ImportError:
    pass
try:
    from rerankers.models.api_rankers import APIRanker

    AVAILABLE_RANKERS["APIRanker"] = APIRanker
except ImportError:
    pass
try:
    from rerankers.models.rankgpt_rankers import RankGPTRanker

    AVAILABLE_RANKERS["RankGPTRanker"] = RankGPTRanker
except ImportError:
    pass
try:
    from rerankers.models.t5ranker import T5Ranker

    AVAILABLE_RANKERS["T5Ranker"] = T5Ranker
except ImportError:
    pass

try:
    from rerankers.models.colbert_ranker import ColBERTRanker

    AVAILABLE_RANKERS["ColBERTRanker"] = ColBERTRanker
except ImportError:
    pass

try:
    from rerankers.models.flashrank_ranker import FlashRankRanker

    AVAILABLE_RANKERS["FlashRankRanker"] = FlashRankRanker
except ImportError:
    pass

try:
    from rerankers.models.rankllm_ranker import RankLLMRanker

    AVAILABLE_RANKERS["RankLLMRanker"] = RankLLMRanker
except ImportError:
    pass



from typing import Optional
import warnings
from rerankers.models import AVAILABLE_RANKERS
from rerankers.models.ranker import BaseRanker
from rerankers.utils import vprint

DEFAULTS = {
    "jina": {"en": "jina-reranker-v1-base-en"},
    "cohere": {"en": "rerank-english-v3.0", "other": "rerank-multilingual-v3.0"},
    "voyage": {"en": "rerank-lite-1"},
    "mixedbread.ai": {"en": "mixedbread-ai/mxbai-rerank-large-v1"},
    "cross-encoder": {
        "en": "mixedbread-ai/mxbai-rerank-base-v1",
        "fr": "antoinelouis/crossencoder-camembert-base-mmarcoFR",
        "zh": "BAAI/bge-reranker-base",
        "other": "corrius/cross-encoder-mmarco-mMiniLMv2-L12-H384-v1",
    },
    "t5": {"en": "unicamp-dl/InRanker-base", "other": "unicamp-dl/mt5-base-mmarco-v2"},
    "lit5": {
        "en": "castorini/LiT5-Distill-base",
    },
    "rankgpt": {"en": "gpt-4-turbo-preview", "other": "gpt-4-turbo-preview"},
    "rankgpt3": {"en": "gpt-3.5-turbo", "other": "gpt-3.5-turbo"},
    "rankgpt4": {"en": "gpt-4", "other": "gpt-4"},
    "rankllm": {"en": "gpt-4o", "other": "gpt-4o"},
    "colbert": {
        "en": "colbert-ir/colbertv2.0",
        "fr": "bclavie/FraColBERTv2",
        "ja": "bclavie/JaColBERTv2",
        "es": "AdrienB134/ColBERTv2.0-spanish-mmarcoES",
    },
    "flashrank": {"en": "ms-marco-MiniLM-L-12-v2", "other": "ms-marco-MultiBERT-L-12"},
    "text-embeddings-inference": {"other": "BAAI/bge-reranker-base"}
}

DEPS_MAPPING = {
    "TransformerRanker": "transformers",
    "T5Ranker": "transformers",
    "LiT5Ranker": "lit5",
    "RankGPTRanker": "gpt",
    "APIRanker": "api",
    "ColBERTRanker": "transformers",
    "FlashRankRanker": "flashrank",
    "RankLLMRanker": "rankllm",
}

PROVIDERS = ["cohere", "jina", "voyage", "mixedbread.ai", "text-embeddings-inference"]


def _get_api_provider(model_name: str, model_type: Optional[str] = None) -> str:
    if model_type in PROVIDERS or any(provider in model_name for provider in PROVIDERS):
        return model_type or next(
            (provider for provider in PROVIDERS if provider in model_name), None
        )
    # Check if the model_name is a key in DEFAULTS to set the provider correctly
    return next(
        (
            provider
            for provider in PROVIDERS
            if model_name in DEFAULTS
            and any(provider in values for values in DEFAULTS[model_name].values())
        ),
        None,
    )


def _get_model_type(model_name: str, explicit_model_type: Optional[str] = None) -> str:
    if explicit_model_type:
        model_mapping = {
            "cohere": "APIRanker",
            "jina": "APIRanker",
            "voyage": "APIRanker",
            "text-embeddings-inference": "APIRanker",
            "rankgpt": "RankGPTRanker",
            "lit5": "LiT5Ranker",
            "t5": "T5Ranker",
            "colbert": "ColBERTRanker",
            "cross-encoder": "TransformerRanker",
            "flashrank": "FlashRankRanker",
            "rankllm": "RankLLMRanker",
        }
        return model_mapping.get(explicit_model_type, explicit_model_type)
    else:
        model_name = model_name.lower()
        model_mapping = {
            "lit5": "LiT5Ranker",
            "t5": "T5Ranker",
            "inranker": "T5Ranker",
            "rankllm": "RankLLMRanker",
            "rankgpt": "RankGPTRanker",
            "gpt": "RankGPTRanker",
            "zephyr": "RankZephyr",
            "colbert": "ColBERTRanker",
            "cohere": "APIRanker",
            "jina": "APIRanker",
            "voyage": "APIRanker",
            "text-embeddings-inference": "APIRanker",
            "ms-marco-minilm-l-12-v2": "FlashRankRanker",
            "ms-marco-multibert-l-12": "FlashRankRanker",
            "vicuna": "RankLLMRanker",
            "zephyr": "RankLLMRanker",
        }
        for key, value in model_mapping.items():
            if key in model_name:
                if key == "gpt":
                    warnings.warn(
                        "The key 'gpt' currently defaults to the rough rankGPT implementation. From version 0.0.5 onwards, 'gpt' will default to RankLLM instead. Please specify the 'rankgpt' `model_type` if you want to keep the current behaviour",
                        DeprecationWarning,
                    )
                return value
        if (
            any(
                keyword in model_name
                for keyword in ["minilm", "bert", "cross-encoders/"]
            )
            and "/" in model_name
        ):
            return "TransformerRanker"
        print(
            "Warning: Model type could not be auto-mapped with the defaults list. Defaulting to TransformerRanker."
        )
        print(
            "If your model is NOT intended to be ran as a one-label cross-encoder, please reload it and specify the model_type!",
            "Otherwise, you may ignore this warning. You may specify `model_type='cross-encoder'` to suppress this warning in the future.",
        )
        return "TransformerRanker"


def _get_defaults(
    model_name: str,
    model_type: Optional[str] = None,
    lang: str = "en",
    verbose: int = 1,
) -> str:
    if model_name in DEFAULTS.keys():
        print(f"Loading default {model_name} model for language {lang}")
        try:
            model_name = DEFAULTS[model_name][lang]
        except KeyError:
            if "other" not in DEFAULTS[model_name]:
                print(
                    f"Model family {model_name} does not have a default for language {lang}"
                )
                print(
                    "Aborting now... Please retry with another model family or by specifying a model"
                )
                return None, None
            model_name = DEFAULTS[model_name]["other"]
    model_type = _get_model_type(model_name, model_type)
    vprint(f"Default Model: {model_name}", verbose)

    return model_name, model_type


def Reranker(
    model_name: str,
    lang: str = "en",
    model_type: Optional[str] = None,
    verbose: int = 1,
    **kwargs,
) -> Optional[BaseRanker]:
    original_model_name = model_name
    api_provider = _get_api_provider(model_name, model_type)
    if api_provider or model_name.lower() in PROVIDERS:
        if model_name.lower() in PROVIDERS:
            api_provider = model_name.lower()
            model_type = "APIRanker"
            model_name = (
                DEFAULTS[api_provider][lang]
                if lang in DEFAULTS[api_provider]
                else DEFAULTS[api_provider]["other"]
            )
            print(
                f"Auto-updated model_name to {model_name} for API provider {api_provider}"
            )
        else:
            model_type = "APIRanker"
    else:
        if original_model_name in DEFAULTS.keys():
            model_name, model_type = _get_defaults(
                original_model_name, model_type, lang, verbose
            )
            if model_name is None:
                return None
            api_provider = _get_api_provider(model_name, model_type)
            if api_provider:
                model_type = "APIRanker"

    if api_provider:
        kwargs["api_provider"] = api_provider

    model_type = _get_model_type(model_name, model_type)

    try:
        print(f"Loading {model_type} model {model_name}")
        return AVAILABLE_RANKERS[model_type](model_name, verbose=verbose, **kwargs)
    except KeyError:
        print(
            f"You don't have the necessary dependencies installed to use {model_type}."
        )
        print(
            f'Please install the necessary dependencies for {model_type} by running `pip install "rerankers[{DEPS_MAPPING[model_type]}]"`',
            'or `pip install "rerankers[all]" to install the dependencies for all reranker types.',
        )
        return None



"""Full implementation is from the original RankGPT repository https://github.com/sunnweiwei/RankGPT under its Apache 2.0 License

Changes made are:
- Truncating the file to only the relevant functions
- Using only LiteLLM
- make_item() added
- Packaging it onto RankGPTRanker"""

import copy
from typing import Optional, Union, List, Dict
from litellm import completion
from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document
from rerankers.results import RankedResults, Result
from rerankers.utils import vprint, prep_docs


def get_prefix_prompt(query, num):
    return [
        {
            "role": "system",
            "content": "You are RankGPT, an intelligent assistant that can rank passages based on their relevancy to the query.",
        },
        {
            "role": "user",
            "content": f"I will provide you with {num} passages, each indicated by number identifier []. \nRank the passages based on their relevance to query: {query}.",
        },
        {"role": "assistant", "content": "Okay, please provide the passages."},
    ]


def get_post_prompt(query, num):
    return f"Search Query: {query}. \nRank the {num} passages above based on their relevance to the search query. The passages should be listed in descending order using identifiers. The most relevant passages should be listed first. The output format should be [] > [], e.g., [1] > [2]. Only response the ranking results, do not say any word or explain."


def create_permutation_instruction(
    item=None,
    rank_start=0,
    rank_end=100,
    lang: str = "en",
):
    query = item["query"]
    num = len(item["hits"][rank_start:rank_end])

    max_length = 300

    messages = get_prefix_prompt(query, num)
    rank = 0
    for hit in item["hits"][rank_start:rank_end]:
        rank += 1
        content = hit["content"]
        content = content.replace("Title: Content: ", "")
        content = content.strip()
        # For Japanese should cut by character: content = content[:int(max_length)]
        if lang in ["zh", "ja"]:
            content = content[: int(max_length)]
        else:
            content = " ".join(content.split()[: int(max_length)])
        messages.append({"role": "user", "content": f"[{rank}] {content}"})
        messages.append({"role": "assistant", "content": f"Received passage [{rank}]."})
    messages.append({"role": "user", "content": get_post_prompt(query, num)})

    return messages


def clean_response(response: str):
    new_response = ""
    for c in response:
        if not c.isdigit():
            new_response += " "
        else:
            new_response += c
    new_response = new_response.strip()
    return new_response


def remove_duplicate(response):
    new_response = []
    for c in response:
        if c not in new_response:
            new_response.append(c)
    return new_response


def receive_permutation(item, permutation, rank_start=0, rank_end=100):
    response = clean_response(permutation)
    response = [int(x) - 1 for x in response.split()]
    response = remove_duplicate(response)
    cut_range = copy.deepcopy(item["hits"][rank_start:rank_end])
    original_rank = [tt for tt in range(len(cut_range))]
    response = [ss for ss in response if ss in original_rank]
    response = response + [tt for tt in original_rank if tt not in response]
    for j, x in enumerate(response):
        item["hits"][j + rank_start] = copy.deepcopy(cut_range[x])
        if "rank" in item["hits"][j + rank_start]:
            item["hits"][j + rank_start]["rank"] = cut_range[j]["rank"]
        if "score" in item["hits"][j + rank_start]:
            item["hits"][j + rank_start]["score"] = cut_range[j]["score"]
    return item


def make_item(
    query: str, docs: List[str]
) -> Dict[str, Union[List[Dict[str, str]], str]]:
    return {
        "query": query,
        "hits": [{"content": doc} for doc in docs],
    }


class RankGPTRanker(BaseRanker):
    def __init__(
        self, model: str, api_key: str, lang: str = "en", verbose: int = 1
    ) -> "RankGPTRanker":
        self.api_key = api_key
        self.model = model
        self.verbose = verbose
        self.lang = lang

    def _query_llm(self, messages: List[Dict[str, str]]) -> str:
        response = completion(
            api_key=self.api_key, model=self.model, messages=messages, temperature=0
        )
        return response.choices[0].message.content

    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
        rank_start: int = 0,
        rank_end: int = 0,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)

        item = make_item(query, [d.text for d in docs])
        messages = create_permutation_instruction(
            item=item,
            rank_start=rank_start,
            rank_end=rank_end,
            lang=self.lang,
        )
        vprint(f"Querying model {self.model} with via LiteLLM...", self.verbose)
        permutation = self._query_llm(messages)
        item = receive_permutation(
            item, permutation, rank_start=rank_start, rank_end=rank_end
        )
        ranked_docs = []
        for idx, doc in enumerate(item["hits"]):
            ranked_docs.append(
                Result(
                    document=list(filter(lambda x: x.text == doc["content"], docs))[0],
                    rank=idx + 1,
                )
            )
        ranked_results = RankedResults(
            results=ranked_docs, query=query, has_scores=False
        )
        return ranked_results

    def score(self):
        print("Listwise ranking models like RankGPT-4 cannot output scores!")
        return None



from typing import Any, Optional, Sequence

from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain_core.callbacks.manager import Callbacks
from langchain_core.documents import Document


class RerankerLangChainCompressor(BaseDocumentCompressor):
    model: Any
    kwargs: dict = {}
    k: int = 5

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,  # noqa
        **kwargs,
    ) -> Any:
        """Rerank a list of documents relevant to a query."""
        doc_list = list(documents)
        _docs = [d.page_content for d in doc_list]
        results = self.model.rank(
            query=query,
            docs=_docs,
            **self.kwargs,
        )
        final_results = []
        for r in results.top_k(kwargs.get("k", self.k)):
            doc = doc_list[r.doc_id]
            doc.metadata["relevance_score"] = r.score
            final_results.append(doc)
        return final_results



from typing import Union, Optional, List
from pydantic import BaseModel, validator

from rerankers.documents import Document


class Result(BaseModel):
    document: Document
    score: Optional[float] = None
    rank: Optional[int] = None

    @validator("rank", always=True)
    def check_score_or_rank_exists(cls, v, values):
        if v is None and values.get("score") is None:
            raise ValueError("Either score or rank must be provided.")
        return v

    def __getattr__(self, item):
        if item in self.document.model_fields:
            return getattr(self.document, item)
        elif item in self.model_fields:
            return getattr(self, item)
        elif item in self.document.metadata:
            return self.document.metadata[item]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{item}'"
        )


class RankedResults(BaseModel):
    results: List[Result]
    query: str
    has_scores: bool = False

    def __iter__(self):
        """Allows iteration over the results list."""
        return iter(self.results)

    def results_count(self) -> int:
        """Returns the total number of results."""
        return len(self.results)

    def top_k(self, k: int) -> List[Result]:
        """Returns the top k results based on the score, if available, or rank."""
        if self.has_scores:
            return sorted(
                self.results,
                key=lambda x: x.score if x.score is not None else float("-inf"),
                reverse=True,
            )[:k]
        else:
            return sorted(
                self.results,
                key=lambda x: x.rank if x.rank is not None else float("inf"),
            )[:k]

    def get_score_by_docid(self, doc_id: Union[int, str]) -> Optional[float]:
        """Fetches the score of a result by its doc_id using a more efficient approach."""
        result = next((r for r in self.results if r.document.doc_id == doc_id), None)
        return result.score if result else None
